
from __future__ import absolute_import
#overrides
import torch
from torch.nn.utils.rnn import pad_packed_sequence

from allennlp.common.checks import ConfigurationError
from allennlp.modules.seq2seq_encoders.seq2seq_encoder import Seq2SeqEncoder


class PytorchSeq2SeqWrapper(Seq2SeqEncoder):
    u"""
    Pytorch's RNNs have two outputs: the hidden state for every time step, and the hidden state at
    the last time step for every layer.  We just want the first one as a single output.  This
    wrapper pulls out that output, and adds a :func:`get_output_dim` method, which is useful if you
    want to, e.g., define a linear + softmax layer on top of this to get some distribution over a
    set of labels.  The linear layer needs to know its input dimension before it is called, and you
    can get that from ``get_output_dim``.

    In order to be wrapped with this wrapper, a class must have the following members:

        - ``self.input_size: int``
        - ``self.hidden_size: int``
        - ``def forward(inputs: PackedSequence, hidden_state: torch.Tensor) ->
          Tuple[PackedSequence, torch.Tensor]``.
        - ``self.bidirectional: bool`` (optional)

    This is what pytorch's RNN's look like - just make sure your class looks like those, and it
    should work.

    Note that we *require* you to pass a binary mask of shape (batch_size, sequence_length)
    when you call this module, to avoid subtle bugs around masking.  If you already have a
    ``PackedSequence`` you can pass ``None`` as the second parameter.

    We support stateful RNNs where the final state from each batch is used as the initial
    state for the subsequent batch by passing ``stateful=True`` to the constructor.
    """
    def __init__(self, module                 , stateful       = False)        :
        super(PytorchSeq2SeqWrapper, self).__init__(stateful)
        self._module = module
        try:
            if not self._module.batch_first:
                raise ConfigurationError(u"Our encoder semantics assumes batch is always first!")
        except AttributeError:
            pass

        try:
            self._is_bidirectional = self._module.bidirectional
        except AttributeError:
            self._is_bidirectional = False
        if self._is_bidirectional:
            self._num_directions = 2
        else:
            self._num_directions = 1

    #overrides
    def get_input_dim(self)       :
        return self._module.input_size

    #overrides
    def get_output_dim(self)       :
        return self._module.hidden_size * self._num_directions

    #overrides
    def is_bidirectional(self)        :
        return self._is_bidirectional

    #overrides
    def forward(self,  # pylint: disable=arguments-differ
                inputs              ,
                mask              ,
                hidden_state               = None)                :

        if self.stateful and mask is None:
            raise ValueError(u"Always pass a mask with stateful RNNs.")
        if self.stateful and hidden_state is not None:
            raise ValueError(u"Stateful RNNs provide their own initial hidden_state.")

        if mask is None:
            return self._module(inputs, hidden_state)[0]

        batch_size, total_sequence_length = mask.size()

        packed_sequence_output, final_states, restoration_indices =\
            self.sort_and_run_forward(self._module, inputs, mask, hidden_state)

        unpacked_sequence_tensor, _ = pad_packed_sequence(packed_sequence_output, batch_first=True)

        num_valid = unpacked_sequence_tensor.size(0)
        # Some RNNs (GRUs) only return one state as a Tensor.  Others (LSTMs) return two.
        # If one state, use a single element list to handle in a consistent manner below.
        if not isinstance(final_states, (list, tuple)) and self.stateful:
            final_states = [final_states]

        # Add back invalid rows.
        if num_valid < batch_size:
            _, length, output_dim = unpacked_sequence_tensor.size()
            zeros = unpacked_sequence_tensor.new_zeros(batch_size - num_valid, length, output_dim)
            unpacked_sequence_tensor = torch.cat([unpacked_sequence_tensor, zeros], 0)

            # The states also need to have invalid rows added back.
            if self.stateful:
                new_states = []
                for state in final_states:
                    num_layers, _, state_dim = state.size()
                    zeros = state.new_zeros(num_layers, batch_size - num_valid, state_dim)
                    new_states.append(torch.cat([state, zeros], 1))
                final_states = new_states

        # It's possible to need to pass sequences which are padded to longer than the
        # max length of the sequence to a Seq2SeqEncoder. However, packing and unpacking
        # the sequences mean that the returned tensor won't include these dimensions, because
        # the RNN did not need to process them. We add them back on in the form of zeros here.
        sequence_length_difference = total_sequence_length - unpacked_sequence_tensor.size(1)
        if sequence_length_difference > 0:
            zeros = unpacked_sequence_tensor.new_zeros(batch_size,
                                                       sequence_length_difference,
                                                       unpacked_sequence_tensor.size(-1))
            unpacked_sequence_tensor = torch.cat([unpacked_sequence_tensor, zeros], 1)

        if self.stateful:
            self._update_states(final_states, restoration_indices)

        # Restore the original indices and return the sequence.
        return unpacked_sequence_tensor.index_select(0, restoration_indices)



from __future__ import print_function

# some python 2 and 3 comnpatibility tweaks
import sys
py3=sys.version_info >= (3, 0)
def inext(v):  # next value from iterator
    return next(v) if py3 else v.next()


import os
import time
import apsw

###
### Check we have the expected version of apsw and sqlite
###

#@@CAPTURE
print ("      Using APSW file",apsw.__file__)                # from the extension module
print ("         APSW version",apsw.apswversion())           # from the extension module
print ("   SQLite lib version",apsw.sqlitelibversion())      # from the sqlite library code
print ("SQLite header version",apsw.SQLITE_VERSION_NUMBER)   # from the sqlite header file at compile time
#@@ENDCAPTURE

###
### Opening/creating database
###

connection=apsw.Connection("dbfile")
cursor=connection.cursor()

###
### simple statement  @@ example-cursor
###

cursor.execute("create table foo(x,y,z)")

###
### using different types
###

cursor.execute("insert into foo values(?,?,?)", (1, 1.1, None))  # integer, float/real, Null
cursor.execute("insert into foo(x) values(?)", ("abc", ))        # string (note trailing comma to ensure tuple!)
cursor.execute("insert into foo(x) values(?)",                   # a blob (binary data)
                    (b"abc\xff\xfe" if py3 else buffer("abc\xff\xfe"), ))

###
### multiple statements
###

cursor.execute("delete from foo; insert into foo values(1,2,3); create table bar(a,b,c) ; insert into foo values(4, 'five', 6.0)")

###
### iterator
###

for x,y,z in cursor.execute("select x,y,z from foo"):
    print (cursor.getdescription())  # shows column names and declared types
    print (x,y,z)

###
### iterator - multiple statements
###

for m,n,o in cursor.execute("select x,y,z from foo ; select a,b,c from bar"):
    print (m,n,o)

###
### bindings - sequence
###

cursor.execute("insert into foo values(?,?,?)", (7, 'eight', False))
cursor.execute("insert into foo values(?,?,?1)", ('one', 'two'))  # nb sqlite does the numbers from 1

###
### bindings - dictionary
###

cursor.execute("insert into foo values(:alpha, :beta, :gamma)", {'alpha': 1, 'beta': 2, 'gamma': 'three'})

###
### tracing execution @@ example-exectrace
###

def mytrace(cursor, statement, bindings):
    "Called just before executing each statement"
    print ("SQL:",statement)
    if bindings:
        print ("Bindings:",bindings)
    return True  # if you return False then execution is aborted

#@@CAPTURE
cursor.setexectrace(mytrace)
cursor.execute("drop table bar ; create table bar(x,y,z); select * from foo where x=?", (3,))
#@@ENDCAPTURE

###
### tracing results @@ example-rowtrace
###

def rowtrace(cursor, row):
    """Called with each row of results before they are handed off.  You can return None to
    cause the row to be skipped or a different set of values to return"""
    print ("Row:", row)
    return row

#@@CAPTURE
cursor.setrowtrace(rowtrace)
for row in cursor.execute("select x,y from foo where x>3"):
     pass
#@@ENDCAPTURE

# Clear tracers
cursor.setrowtrace(None)
cursor.setexectrace(None)

###
### executemany
###

# (This will work correctly with multiple statements, as well as statements that
# return data.  The second argument can be anything that is iterable.)
cursor.executemany("insert into foo (x) values(?)", ( [1], [2], [3] ) )

# You can also use it for statements that return data
for row in cursor.executemany("select * from foo where x=?", ( [1], [2], [3] ) ):
    print (row)

###
### defining your own functions  @@ scalar-example
###

def ilove7(*args):
    "a scalar function"
    print ("ilove7 got",args,"but I love 7")
    return 7

connection.createscalarfunction("seven", ilove7)

#@@CAPTURE
for row in cursor.execute("select seven(x,y) from foo"):
    print (row)
#@@ENDCAPTURE

###
### aggregate functions are more complex @@ aggregate-example
###

# Here we return the longest item when represented as a string.

class longest:
    def __init__(self):
        self.longest=""

    def step(self, *args):
        for arg in args:
            if len( str(arg) ) > len (self.longest):
                self.longest=str(arg)

    def final(self):
        return self.longest

    # Under Python 2.3 remove the following line and add
    # factory=classmethod(factory) at the end
    @classmethod
    def factory(cls):
        return cls(), cls.step, cls.final

#@@CAPTURE
connection.createaggregatefunction("longest", longest.factory)
for row in cursor.execute("select longest(x,y) from foo"):
    print (row)
#@@ENDCAPTURE

###
### Defining collations.  @@ collation-example
###

# The default sorting mechanisms don't understand numbers at the end of strings
# so here we define a collation that does

cursor.execute("create table s(str)")
cursor.executemany("insert into s values(?)",
                  ( ["file1"], ["file7"], ["file17"], ["file20"], ["file3"] ) )

#@@CAPTURE
for row in cursor.execute("select * from s order by str"):
    print (row)
#@@ENDCAPTURE

def strnumcollate(s1, s2):
    # return -1 if s1<s2, +1 if s1>s2 else 0

    # split values into two parts - the head and the numeric tail
    values=[s1, s2]
    for vn,v in enumerate(values):
        for i in range(len(v), 0, -1):
            if v[i-1] not in "01234567890":
                break
        try:
            v=( v[:i], int(v[i:]) )
        except ValueError:
            v=( v[:i], None )
        values[vn]=v
    # compare
    if values[0]<values[1]:
        return -1
    if values[0]>values[1]:
        return 1
    return 0

connection.createcollation("strnum", strnumcollate)

#@@CAPTURE
for row in cursor.execute("select * from s order by str collate strnum"):
    print (row)
#@@ENDCAPTURE

###
### Authorizer (eg if you want to control what user supplied SQL can do) @@ authorizer-example
###

def authorizer(operation, paramone, paramtwo, databasename, triggerorview):
    """Called when each operation is prepared.  We can return SQLITE_OK, SQLITE_DENY or
    SQLITE_IGNORE"""
    # find the operation name
    print (apsw.mapping_authorizer_function[operation], paramone, paramtwo, databasename, triggerorview)
    if operation==apsw.SQLITE_CREATE_TABLE and paramone.startswith("private"):
        return apsw.SQLITE_DENY  # not allowed to create tables whose names start with private

    return apsw.SQLITE_OK  # always allow

connection.setauthorizer(authorizer)
#@@CAPTURE
cursor.execute("insert into s values('foo')")
cursor.execute("select str from s limit 1")
#@@ENDCAPTURE

# Cancel authorizer
connection.setauthorizer(None)

###
### progress handler (SQLite 3 experimental feature) @@ example-progress-handler
###

# something to give us large numbers of random numbers
import random
def randomintegers(howmany):
    for i in range(howmany):
        yield (random.randint(0,9999999999),)

# create a table with 100 random numbers
cursor.execute("begin ; create table bigone(x)")
cursor.executemany("insert into bigone values(?)", randomintegers(100))
cursor.execute("commit")

# display an ascii spinner
_phcount=0
_phspinner="|/-\\"
def progresshandler():
    global _phcount
    sys.stdout.write(_phspinner[_phcount%len(_phspinner)]+chr(8)) # chr(8) is backspace
    sys.stdout.flush()
    _phcount+=1
    time.sleep(0.1) # deliberate delay so we can see the spinner (SQLite is too fast otherwise!)
    return 0  # returning non-zero aborts

# register progresshandler every 20 instructions
connection.setprogresshandler(progresshandler, 20)

# see it in action - sorting 100 numbers to find the biggest takes a while
print ("spinny thing -> ", end="")
for i in cursor.execute("select max(x) from bigone"):
    print("\n", i, sep="", end="")
    sys.stdout.flush()

connection.setprogresshandler(None)

###
### commit hook (SQLite3 experimental feature) @@ example-commithook
###

def mycommithook():
    print ("in commit hook")
    hour=time.localtime()[3]
    if hour<8 or hour>17:
        print ("no commits out of hours")
        return 1  # abort commits outside of 8am through 6pm
    print ("commits okay at this time")
    return 0  # let commit go ahead

#@@CAPTURE
connection.setcommithook(mycommithook)
try:
    cursor.execute("begin; create table example(x,y,z); insert into example values (3,4,5) ; commit")
except apsw.ConstraintError:
    print ("commit was not allowed")

connection.setcommithook(None)
#@@ENDCAPTURE

###
### update hook @@ example-updatehook
###

def myupdatehook(type, databasename, tablename, rowid):
    print ("Updated: %s database %s, table %s, row %d" % (
        apsw.mapping_authorizer_function[type], databasename, tablename, rowid))

#@@CAPTURE
connection.setupdatehook(myupdatehook)
cursor.execute("insert into s values(?)", ("file93",))
cursor.execute("update s set str=? where str=?", ("file94", "file93"))
cursor.execute("delete from s where str=?", ("file94",))
connection.setupdatehook(None)
#@@ENDCAPTURE

###
### Blob I/O    @@ example-blobio
###

cursor.execute("create table blobby(x,y)")
# Add a blob we will fill in later
cursor.execute("insert into blobby values(1,zeroblob(10000))")
# Or as a binding
cursor.execute("insert into blobby values(2,?)", (apsw.zeroblob(20000),))
# Open a blob for writing.  We need to know the rowid
rowid=inext(cursor.execute("select ROWID from blobby where x=1"))[0]
blob=connection.blobopen("main", "blobby", "y", rowid, 1) # 1 is for read/write
blob.write(b"hello world")
blob.seek(2000)
blob.write(b"hello world, again")
blob.close()

###
### Virtual tables  @@ example-vtable
###

# This virtual table stores information about files in a set of
# directories so you can execute SQL queries

def getfiledata(directories):
    columns=None
    data=[]
    counter=1
    for directory in directories:
        for f in os.listdir(directory):
            if not os.path.isfile(os.path.join(directory,f)):
                continue
            counter+=1
            st=os.stat(os.path.join(directory,f))
            if columns is None:
                columns=["rowid", "name", "directory"]+[x for x in dir(st) if x.startswith("st_")]
            data.append( [counter, f, directory] + [getattr(st,x) for x in columns[3:]] )
    return columns, data

# This gets registered with the Connection
class Source:
    def Create(self, db, modulename, dbname, tablename, *args):
        columns,data=getfiledata([eval(a.replace("\\", "\\\\")) for a in args]) # eval strips off layer of quotes
        schema="create table foo("+','.join(["'%s'" % (x,) for x in columns[1:]])+")"
        return schema,Table(columns,data)
    Connect=Create

# Represents a table
class Table:
    def __init__(self, columns, data):
        self.columns=columns
        self.data=data

    def BestIndex(self, *args):
        return None

    def Open(self):
        return Cursor(self)

    def Disconnect(self):
        pass

    Destroy=Disconnect

# Represents a cursor
class Cursor:
    def __init__(self, table):
        self.table=table

    def Filter(self, *args):
        self.pos=0

    def Eof(self):
        return self.pos>=len(self.table.data)

    def Rowid(self):
        return self.table.data[self.pos][0]

    def Column(self, col):
        return self.table.data[self.pos][1+col]

    def Next(self):
        self.pos+=1

    def Close(self):
        pass

# Register the module as filesource
connection.createmodule("filesource", Source())

# Arguments to module - all directories in sys.path
sysdirs=",".join(["'%s'" % (x,) for x in sys.path[1:] if len(x) and os.path.isdir(x)])
cursor.execute("create virtual table sysfiles using filesource("+sysdirs+")")

#@@CAPTURE
# Which 3 files are the biggest?
for size,directory,file in cursor.execute("select st_size,directory,name from sysfiles order by st_size desc limit 3"):
    print (size,file,directory)
#@@ENDCAPTURE

# Which 3 files are the oldest?
#@@CAPTURE
for ctime,directory,file in cursor.execute("select st_ctime,directory,name from sysfiles order by st_ctime limit 3"):
    print (ctime,file,directory)
#@@ENDCAPTURE

### @@ example-vfs
### A VFS that "obfuscates" the database file contents.  The scheme
### used is to xor all bytes with 0xa5.  This scheme honours that used
### for MAPI and SQL Server.
###

def encryptme(data):
    if not data: return data
    if py3:
        return bytes([x^0xa5 for x in data])

    return "".join([chr(ord(x)^0xa5) for x in data])

# Inheriting from a base of "" means the default vfs
class ObfuscatedVFS(apsw.VFS):
    def __init__(self, vfsname="obfu", basevfs=""):
        self.vfsname=vfsname
        self.basevfs=basevfs
        apsw.VFS.__init__(self, self.vfsname, self.basevfs)

    # We want to return our own file implmentation, but also
    # want it to inherit
    def xOpen(self, name, flags):
        # We can look at uri parameters
        if isinstance(name, apsw.URIFilename):
            #@@CAPTURE
            print ("fast is", name.uri_parameter("fast"))
            print ("level is", name.uri_int("level", 3))
            print ("warp is", name.uri_boolean("warp", False))
            print ("notpresent is", name.uri_parameter("notpresent"))
            #@@ENDCAPTURE
        return ObfuscatedVFSFile(self.basevfs, name, flags)

# The file implementation where we override xRead and xWrite to call our
# encryption routine
class ObfuscatedVFSFile(apsw.VFSFile):
    def __init__(self, inheritfromvfsname, filename, flags):
        apsw.VFSFile.__init__(self, inheritfromvfsname, filename, flags)

    def xRead(self, amount, offset):
        return encryptme(super(ObfuscatedVFSFile, self).xRead(amount, offset))

    def xWrite(self, data, offset):
        super(ObfuscatedVFSFile, self).xWrite(encryptme(data), offset)

# To register the VFS we just instantiate it
obfuvfs=ObfuscatedVFS()
# Lets see what vfs are now available?
#@@CAPTURE
print (apsw.vfsnames())
#@@ENDCAPTURE

# Make an obfuscated db, passing in some URI parameters
obfudb=apsw.Connection("file:myobfudb?fast=speed&level=7&warp=on",
                       flags=apsw.SQLITE_OPEN_READWRITE | apsw.SQLITE_OPEN_CREATE | apsw.SQLITE_OPEN_URI,
                       vfs=obfuvfs.vfsname)
# Check it works
obfudb.cursor().execute("create table foo(x,y); insert into foo values(1,2)")

# Check it really is obfuscated on disk
#@@CAPTURE
print (open("myobfudb", "rb").read()[:20])
#@@ENDCAPTURE

# And unobfuscating it
#@@CAPTURE
print (encryptme(open("myobfudb", "rb").read()[:20]))
#@@ENDCAPTURE

# Tidy up
obfudb.close()
os.remove("myobfudb")


###
### Limits @@example-limit
###

#@@CAPTURE
# Print some limits
for limit in ("LENGTH", "COLUMN", "ATTACHED"):
    name="SQLITE_LIMIT_"+limit
    maxname="SQLITE_MAX_"+limit  # compile time
    orig=connection.limit(getattr(apsw, name))
    print (name, orig)
    # To get the maximum, set to 0x7fffffff and then read value back
    connection.limit(getattr(apsw, name), 0x7fffffff)
    max=connection.limit(getattr(apsw, name))
    print (maxname, max)

# Set limit for size of a string
cursor.execute("create table testlimit(s)")
cursor.execute("insert into testlimit values(?)", ( "x"*1024, )) # 1024 char string
connection.limit(apsw.SQLITE_LIMIT_LENGTH, 1023) # limit is now 1023
try:
    cursor.execute("insert into testlimit values(?)", ( "y"*1024, ))
    print ("string exceeding limit was inserted")
except apsw.TooBigError:
    print ("Caught toobig exception")
connection.limit(apsw.SQLITE_LIMIT_LENGTH, 0x7fffffff)

#@@ENDCAPTURE

###
### Backup to memory @@example-backup
###

# We will copy the disk database into a memory database

memcon=apsw.Connection(":memory:")

# Copy into memory
with memcon.backup("main", connection, "main") as backup:
    backup.step() # copy whole database in one go

# There will be no disk accesses for this query
for row in memcon.cursor().execute("select * from s"):
    pass

###
### Shell  @@ example-shell
###

# Here we use the shell to do a csv export providing the existing db
# connection

# Export to a StringIO
if py3:
    import io
else:
    import StringIO as io
    
output=io.StringIO()
shell=apsw.Shell(stdout=output, db=connection)
# How to execute a dot command
shell.process_command(".mode csv")
shell.process_command(".headers on")
# How to execute SQL
shell.process_sql("create table csvtest(col1,col2); insert into csvtest values(3,4); insert into csvtest values('a b', NULL)")
# Let the shell figure out SQL vs dot command
shell.process_complete_line("select * from csvtest")

# Verify output
#@@CAPTURE
print (output.getvalue())
#@@ENDCAPTURE

###
### Statistics @@example-status
###

#@@CAPTURE
print ("SQLite memory usage current %d max %d" % apsw.status(apsw.SQLITE_STATUS_MEMORY_USED))
#@@ENDCAPTURE

###
### Cleanup
###

# We can close connections manually (useful if you want to catch exceptions)
# but you don't have to
connection.close(True)  # force it since we want to exit

# Delete database - we don't need it any more
os.remove("dbfile")



# pylint: disable=no-self-use,invalid-name,protected-access



from __future__ import with_statement
from __future__ import division
from __future__ import absolute_import
import json

#overrides

from allennlp.common.testing import AllenNlpTestCase
from allennlp.data.tokenizers import Token
from allennlp.semparse import ParsingError, World
from allennlp.semparse.contexts import TableQuestionKnowledgeGraph
from allennlp.semparse.worlds import NlvrWorld, WikiTablesWorld
from io import open


class FakeWorldWithoutRecursion(World):
    # pylint: disable=abstract-method
    #overrides
    def all_possible_actions(self):
        # The logical forms this grammar allows are
        # (unary_function argument)
        # (binary_function argument argument)
        actions = [u'@start@ -> t',
                   u't -> [<e,t>, e]',
                   u'<e,t> -> unary_function',
                   u'<e,t> -> [<e,<e,t>>, e]',
                   u'<e,<e,t>> -> binary_function',
                   u'e -> argument']
        return actions


class FakeWorldWithRecursion(FakeWorldWithoutRecursion):
    # pylint: disable=abstract-method
    #overrides
    def all_possible_actions(self):
        # In addition to the forms allowed by ``FakeWorldWithoutRecursion``, this world allows
        # (unary_function (identity .... (argument)))
        # (binary_function (identity .... (argument)) (identity .... (argument)))
        actions = super(FakeWorldWithRecursion, self).all_possible_actions()
        actions.extend([u'e -> [<e,e>, e]',
                        u'<e,e> -> identity'])
        return actions


class TestWorld(AllenNlpTestCase):
    def setUp(self):
        super(TestWorld, self).setUp()
        self.world_without_recursion = FakeWorldWithoutRecursion()
        self.world_with_recursion = FakeWorldWithRecursion()

        test_filename = self.FIXTURES_ROOT / u"data" / u"nlvr" / u"sample_ungrouped_data.jsonl"
        data = [json.loads(line)[u"structured_rep"] for line in open(test_filename).readlines()]
        self.nlvr_world = NlvrWorld(data[0])

        question_tokens = [Token(x) for x in [u'what', u'was', u'the', u'last', u'year', u'2004', u'?']]
        table_file = self.FIXTURES_ROOT / u'data' / u'wikitables' / u'sample_table.tsv'
        table_kg = TableQuestionKnowledgeGraph.read_from_file(table_file, question_tokens)
        self.wikitables_world = WikiTablesWorld(table_kg)

    def test_get_paths_to_root_without_recursion(self):
        argument_paths = self.world_without_recursion.get_paths_to_root(u'e -> argument')
        assert argument_paths == [[u'e -> argument', u't -> [<e,t>, e]', u'@start@ -> t'],
                                  [u'e -> argument', u'<e,t> -> [<e,<e,t>>, e]', u't -> [<e,t>, e]',
                                   u'@start@ -> t']]
        unary_function_paths = self.world_without_recursion.get_paths_to_root(u'<e,t> -> unary_function')
        assert unary_function_paths == [[u'<e,t> -> unary_function', u't -> [<e,t>, e]',
                                         u'@start@ -> t']]
        binary_function_paths =\
                self.world_without_recursion.get_paths_to_root(u'<e,<e,t>> -> binary_function')
        assert binary_function_paths == [[u'<e,<e,t>> -> binary_function',
                                          u'<e,t> -> [<e,<e,t>>, e]', u't -> [<e,t>, e]',
                                          u'@start@ -> t']]

    def test_get_paths_to_root_with_recursion(self):
        argument_paths = self.world_with_recursion.get_paths_to_root(u'e -> argument')
        # Argument now has 4 paths, and the two new paths are with the identity function occurring
        # (only once) within unary and binary functions.
        assert argument_paths == [[u'e -> argument', u't -> [<e,t>, e]', u'@start@ -> t'],
                                  [u'e -> argument', u'<e,t> -> [<e,<e,t>>, e]', u't -> [<e,t>, e]',
                                   u'@start@ -> t'],
                                  [u'e -> argument', u'e -> [<e,e>, e]', u't -> [<e,t>, e]',
                                   u'@start@ -> t'],
                                  [u'e -> argument', u'e -> [<e,e>, e]', u'<e,t> -> [<e,<e,t>>, e]',
                                   u't -> [<e,t>, e]', u'@start@ -> t']]
        identity_paths = self.world_with_recursion.get_paths_to_root(u'<e,e> -> identity')
        # Two identity paths, one through each of unary and binary function productions.
        assert identity_paths == [[u'<e,e> -> identity', u'e -> [<e,e>, e]', u't -> [<e,t>, e]',
                                   u'@start@ -> t'],
                                  [u'<e,e> -> identity', u'e -> [<e,e>, e]',
                                   u'<e,t> -> [<e,<e,t>>, e]', u't -> [<e,t>, e]', u'@start@ -> t']]

    # The tests for get_action_sequence and get_logical_form need a concrete world to be useful;
    # we'll mostly use the NLVR world to test them, as it's a simpler world than the WikiTables
    # world.

    def test_get_action_sequence_removes_currying(self):
        world = self.wikitables_world
        logical_form = (u"(argmax (number 1) (number 1) (fb:row.row.division fb:cell.2) "
                        u"(reverse (lambda x ((reverse fb:row.row.index) (var x))))")
        parsed_logical_form = world.parse_logical_form(logical_form)
        action_sequence = world.get_action_sequence(parsed_logical_form)
        assert u'r -> [<n,<n,<#1,<<#2,#1>,#1>>>>, n, n, r, <n,r>]' in action_sequence

    def test_get_action_sequence_removes_and_retains_var_correctly(self):
        world = self.wikitables_world
        logical_form = (u"((reverse fb:row.row.league) (argmin (number 1) (number 1) "
                        u"(fb:type.object.type fb:type.row) "
                        u"(reverse (lambda x ((reverse fb:row.row.index) (var x))))))")
        parsed_logical_form_without_var = world.parse_logical_form(logical_form)
        action_sequence_without_var = world.get_action_sequence(parsed_logical_form_without_var)
        assert u'<#1,#1> -> var' not in action_sequence_without_var

        parsed_logical_form_with_var = world.parse_logical_form(logical_form,
                                                                remove_var_function=False)
        action_sequence_with_var = world.get_action_sequence(parsed_logical_form_with_var)
        assert u'<#1,#1> -> var' in action_sequence_with_var

    def test_get_logical_form_handles_reverse(self):
        world = self.wikitables_world
        logical_form = u"((reverse fb:row.row.year) (fb:row.row.league fb:cell.usl_a_league))"
        parsed_logical_form = world.parse_logical_form(logical_form)
        action_sequence = world.get_action_sequence(parsed_logical_form)
        reconstructed_logical_form = world.get_logical_form(action_sequence)
        parsed_reconstructed_logical_form = world.parse_logical_form(reconstructed_logical_form)
        assert parsed_logical_form == parsed_reconstructed_logical_form

        logical_form = (u"((reverse fb:cell.cell.date) ((reverse fb:row.row.year) (argmax (number 1) "
                        u"(number 1) (fb:row.row.league fb:cell.usl_a_league) (reverse (lambda x "
                        u"((reverse fb:row.row.index) (var x)))))))")
        parsed_logical_form = world.parse_logical_form(logical_form)
        action_sequence = world.get_action_sequence(parsed_logical_form)
        reconstructed_logical_form = world.get_logical_form(action_sequence)
        parsed_reconstructed_logical_form = world.parse_logical_form(reconstructed_logical_form)
        assert parsed_logical_form == parsed_reconstructed_logical_form

    def test_get_logical_form_handles_greater_than(self):
        world = self.wikitables_world
        action_sequence = [u'@start@ -> c', u'c -> [<r,c>, r]', u'<r,c> -> [<<#1,#2>,<#2,#1>>, <c,r>]',
                           u'<<#1,#2>,<#2,#1>> -> reverse', u'<c,r> -> fb:row.row.league',
                           u'r -> [<c,r>, c]', u'<c,r> -> fb:row.row.year', u'c -> [<n,c>, n]',
                           u'<n,c> -> fb:cell.cell.number', u'n -> [<nd,nd>, n]', u'<nd,nd> -> >',
                           u'n -> [<n,n>, n]', u'<n,n> -> number', u'n -> 2004']
        logical_form = world.get_logical_form(action_sequence)
        expected_logical_form = (u'((reverse fb:row.row.league) (fb:row.row.year '
                                 u'(fb:cell.cell.number (> (number 2004)))))')
        assert logical_form == expected_logical_form

    def test_get_logical_form_handles_length_one_terminal_functions(self):
        world = self.wikitables_world
        logical_form = (u"(- ((reverse fb:cell.cell.number) ((reverse fb:row.row.league) "
                        u"(fb:row.row.year fb:cell.usl_a_league))) (number 1))")
        parsed_logical_form = world.parse_logical_form(logical_form)
        action_sequence = world.get_action_sequence(parsed_logical_form)
        reconstructed_logical_form = world.get_logical_form(action_sequence)
        parsed_reconstructed_logical_form = world.parse_logical_form(reconstructed_logical_form)
        assert parsed_logical_form == parsed_reconstructed_logical_form

    def test_get_logical_form_with_real_logical_forms(self):
        nlvr_world = self.nlvr_world
        logical_form = (u"(box_count_greater_equals (member_color_count_equals all_boxes 1) 1)")
        parsed_logical_form = nlvr_world.parse_logical_form(logical_form)
        action_sequence = nlvr_world.get_action_sequence(parsed_logical_form)
        reconstructed_logical_form = nlvr_world.get_logical_form(action_sequence)
        parsed_reconstructed_logical_form = nlvr_world.parse_logical_form(reconstructed_logical_form)
        # It makes more sense to compare parsed logical forms instead of actual logical forms.
        assert parsed_logical_form == parsed_reconstructed_logical_form
        assert nlvr_world.execute(logical_form) == nlvr_world.execute(reconstructed_logical_form)
        logical_form = u"(object_color_all_equals (circle (touch_wall (all_objects))) color_black)"
        parsed_logical_form = nlvr_world.parse_logical_form(logical_form)
        action_sequence = nlvr_world.get_action_sequence(parsed_logical_form)
        reconstructed_logical_form = nlvr_world.get_logical_form(action_sequence)
        parsed_reconstructed_logical_form = nlvr_world.parse_logical_form(reconstructed_logical_form)
        assert parsed_logical_form == parsed_reconstructed_logical_form
        assert nlvr_world.execute(logical_form) == nlvr_world.execute(reconstructed_logical_form)

    def test_get_logical_form_fails_with_incomplete_action_sequence(self):
        nlvr_world = self.nlvr_world
        action_sequence = [u'@start@ -> t', u't -> [<b,t>, b]', u'<b,t> -> box_exists']
        with self.assertRaisesRegex(ParsingError, u'Incomplete action sequence'):
            nlvr_world.get_logical_form(action_sequence)

    def test_get_logical_form_fails_with_extra_actions(self):
        nlvr_world = self.nlvr_world
        action_sequence = [u'@start@ -> <b,t>', u'<b,t> -> box_exists', u't -> [<b,t>, b]']
        with self.assertRaisesRegex(ParsingError, u'Extra actions'):
            nlvr_world.get_logical_form(action_sequence)

    def test_get_logical_form_fails_with_action_sequence_in_wrong_order(self):
        nlvr_world = self.nlvr_world
        action_sequence = [u'@start@ -> t', u't -> [<b,t>, b]', u'<b,t> -> box_exists',
                           u'b -> [<c,b>, c]', u'<c,b> -> [<b,<c,b>>, b]',
                           u'b -> all_boxes', u'<b,<c,b>> -> member_color_none_equals',
                           u'c -> color_blue']
        with self.assertRaisesRegex(ParsingError, u'does not match'):
            nlvr_world.get_logical_form(action_sequence)

    def test_get_logical_form_adds_var_correctly(self):
        world = self.wikitables_world
        action_sequence = [u'@start@ -> e', u'e -> [<r,e>, r]', u'<r,e> -> [<<#1,#2>,<#2,#1>>, <e,r>]',
                           u'<<#1,#2>,<#2,#1>> -> reverse', u'<e,r> -> fb:row.row.league',
                           u'r -> [<d,<d,<#1,<<d,#1>,#1>>>>, d, d, r, <d,r>]',
                           u'<d,<d,<#1,<<d,#1>,#1>>>> -> argmin', u'd -> [<e,d>, e]', u'<e,d> -> number',
                           u'e -> 1', u'd -> [<e,d>, e]', u'<e,d> -> number', u'e -> 1',
                           u'r -> [<#1,#1>, r]', u'<#1,#1> -> fb:type.object.type', u'r -> fb:type.row',
                           u'<d,r> -> [<<#1,#2>,<#2,#1>>, <r,d>]', u'<<#1,#2>,<#2,#1>> -> reverse',
                           u"<r,d> -> ['lambda x', d]", u'd -> [<r,d>, r]',
                           u'<r,d> -> [<<#1,#2>,<#2,#1>>, <d,r>]', u'<<#1,#2>,<#2,#1>> -> reverse',
                           u'<d,r> -> fb:row.row.index', u'r -> x']
        logical_form = world.get_logical_form(action_sequence)
        assert u'(var x)' in logical_form
        expected_logical_form = (u"((reverse fb:row.row.league) (argmin (number 1) (number 1) "
                                 u"(fb:type.object.type fb:type.row) "
                                 u"(reverse (lambda x ((reverse fb:row.row.index) (var x))))))")
        parsed_logical_form = world.parse_logical_form(logical_form)
        parsed_expected_logical_form = world.parse_logical_form(expected_logical_form)
        assert parsed_logical_form == parsed_expected_logical_form

    def test_get_logical_form_fails_with_unnecessary_add_var(self):
        world = self.wikitables_world
        action_sequence = [u'@start@ -> e', u'e -> [<r,e>, r]', u'<r,e> -> [<<#1,#2>,<#2,#1>>, <e,r>]',
                           u'<<#1,#2>,<#2,#1>> -> reverse', u'<e,r> -> fb:row.row.league',
                           u'r -> [<d,<d,<#1,<<d,#1>,#1>>>>, d, d, r, <d,r>]',
                           u'<d,<d,<#1,<<d,#1>,#1>>>> -> argmin', u'd -> [<e,d>, e]', u'<e,d> -> number',
                           u'e -> 1', u'd -> [<e,d>, e]', u'<e,d> -> number', u'e -> 1',
                           u'r -> [<#1,#1>, r]', u'<#1,#1> -> fb:type.object.type', u'r -> fb:type.row',
                           u'<d,r> -> [<<#1,#2>,<#2,#1>>, <r,d>]', u'<<#1,#2>,<#2,#1>> -> reverse',
                           u"<r,d> -> ['lambda x', d]", u'd -> [<r,d>, r]',
                           u'<r,d> -> [<<#1,#2>,<#2,#1>>, <d,r>]', u'<<#1,#2>,<#2,#1>> -> reverse',
                           u'<d,r> -> fb:row.row.index', u'r -> [<#1,#1>, r]', u'<#1,#1> -> var', u'r -> x']
        with self.assertRaisesRegex(ParsingError, u'already had var'):
            world.get_logical_form(action_sequence)

    def test_get_logical_form_with_multiple_negate_filters(self):
        world = self.nlvr_world
        # This is an actual sequence of actions produced by an untrained NlvrSemanticParser
        action_sequence = [u'@start@ -> t', u't -> [<o,<c,t>>, o, c]',
                           u'<o,<c,t>> -> object_color_all_equals', u'o -> [<o,o>, o]',
                           u'<o,o> -> [<<o,o>,<o,o>>, <o,o>]', u'<<o,o>,<o,o>> -> negate_filter',
                           u'<o,o> -> [<<o,o>,<o,o>>, <o,o>]', u'<<o,o>,<o,o>> -> negate_filter',
                           u'<o,o> -> blue', u'o -> [<o,o>, o]', u'<o,o> -> blue', u'o -> [<o,o>, o]',
                           u'<o,o> -> blue', u'o -> all_objects', u'c -> color_blue']
        logical_form = world.get_logical_form(action_sequence)
        # "The color of all blue blue objects that are not not blue is blue".
        expected_logical_form = (u"(object_color_all_equals ((negate_filter (negate_filter blue)) "
                                 u"(blue (blue all_objects))) color_blue)")
        parsed_logical_form = world.parse_logical_form(logical_form)
        parsed_expected_logical_form = world.parse_logical_form(expected_logical_form)
        assert parsed_logical_form == parsed_expected_logical_form



# Python code
#
# See the accompanying LICENSE file.
#
# The purpose of this file is to produce rst output interspersed into
# the the text of the example code

# Imports
import string, sys, cStringIO, os, re

def indentof(l):
    return l[:len(l)-len(l.lstrip())]

def docapture(filename):
    code=[]
    code.append(outputredirector)
    counter=0
    for line in open(filename, "rU"):
        line=line[:-1] # strip off newline
        if line.lstrip().startswith("#@@CAPTURE"):
            code.append(indentof(line)+"opto('.tmpop-%s-%d')" % (filename, counter))
            counter+=1
        elif line.lstrip().startswith("#@@ENDCAPTURE"):
            code.append(indentof(line)+"opnormal()")
        else:
            code.append(line)
    code="\n".join(code)
    # open("xx.py", "wt").write(code)
    exec code in {}

outputredirector="""
import sys
origsysstdout=None
def opto(fname):
  global origsysstdout
  origsysstdout=sys.stdout,fname
  sys.stdout=open(fname, "wt")
def opnormal():
  sys.stdout.close()
  sys.stdout=origsysstdout[0]
  sys.stdout.write(open(origsysstdout[1], "rb").read())
"""

def rstout(filename):
    op=[]
    op.extend("""
.. Automatically generated by example2rst.py.  Edit that file
   not this one!
   
Example
=======

This code demonstrates usage of the APSW api.  It gives you a good
overview of all the things that can be done.  Also included is output
so you can see what gets printed when you run the code.

.. code-block:: python
""".split("\n"))
    counter=0
    prefix="    "
    for line in open(filename, "rtU"):
        line=line.rstrip()
        if "@@CAPTURE" in line:
            continue
        if "@@ENDCAPTURE" not in line:
            if line.startswith("#") and "@@" in line:
                p=line.index("@@")
                name=line[p+2:].strip()
                line=line[:p].rstrip()
                for i in range(-1,-99,-1): # look backwards for first non-comment line
                    if not op[i].strip().startswith("#"):
                        op.insert(i, "")
                        op.insert(i, ".. _"+name+":")
                        op.insert(i, "")
                        op.insert(i, ".. code-block:: python")
                        op.insert(i, "")
                        break
            op.append(prefix+line)
            continue
        op.append("")
        op.append(".. code-block:: text")
        op.append("")
        for line in open(".tmpop-%s-%d" % (filename, counter), "rtU"):
            line=line.rstrip()
            op.append("   | "+re.sub("u'([^']*)'", r"'\1'", line))
        op.append("")
        op.append(".. code-block:: python")
        op.append("")
        os.remove(".tmpop-%s-%d" % (filename, counter))
        counter+=1

    ### Peephole optimizations

    while True:
        b4=op
        # get rid of double blank lines
        op2=[]
        for i in range(len(op)):
            if i+1<len(op) and len(op[i].strip())==0 and len(op[i+1].strip())==0:
                continue
            if len(op[i].strip())==0: # make whitespace only lines be zero length
                op2.append("")
            else:
                op2.append(op[i])
        op=op2

        # if there is a code block followed by a label then drop the code block
        op2=[]
        for i in range(len(op)):
            if i+2<len(op) and op[i].startswith(".. code-block::") and op[i+2].startswith(".. _"):
                continue
            op2.append(op[i])
        op=op2
        
        if op==b4:
            break


    return op

if __name__ == "__main__":
  docapture("example-code.py")
  op=rstout("example-code.py")
  open("doc/example.rst", "wt").write("\n".join(op)+"\n")



# pylint: disable=no-self-use,invalid-name



from __future__ import with_statement
from __future__ import division
from __future__ import absolute_import
import filecmp
import json
import os
import pathlib
import tarfile

import torch

from allennlp.commands.train import train_model
from allennlp.common import Params
from allennlp.common.testing import ModelTestCase
from allennlp.data.dataset import Batch
from allennlp.modules.token_embedders import ElmoTokenEmbedder
from io import open


class TestElmoTokenEmbedder(ModelTestCase):
    def setUp(self):
        super(TestElmoTokenEmbedder, self).setUp()
        self.set_up_model(self.FIXTURES_ROOT / u'elmo' / u'config' / u'characters_token_embedder.json',
                          self.FIXTURES_ROOT / u'data' / u'conll2003.txt')

    def test_tagger_with_elmo_token_embedder_can_train_save_and_load(self):
        self.ensure_model_can_train_save_and_load(self.param_file)

    def test_tagger_with_elmo_token_embedder_forward_pass_runs_correctly(self):
        dataset = Batch(self.instances)
        dataset.index_instances(self.vocab)
        training_tensors = dataset.as_tensor_dict()
        output_dict = self.model(**training_tensors)
        tags = output_dict[u'tags']
        assert len(tags) == 2
        assert len(tags[0]) == 7
        assert len(tags[1]) == 7
        for example_tags in tags:
            for tag_id in example_tags:
                tag = self.model.vocab.get_token_from_index(tag_id, namespace=u"labels")
                assert tag in set([u'O', u'I-ORG', u'I-PER', u'I-LOC'])

    def test_file_archiving(self):
        # This happens to be a good place to test auxiliary file archiving.
        # Train the model
        params = Params.from_file(self.FIXTURES_ROOT / u'elmo' / u'config' / u'characters_token_embedder.json')
        serialization_dir = os.path.join(self.TEST_DIR, u'serialization')
        train_model(params, serialization_dir)

        # Inspect the archive
        archive_file = os.path.join(serialization_dir, u'model.tar.gz')
        unarchive_dir = os.path.join(self.TEST_DIR, u'unarchive')
        with tarfile.open(archive_file, u'r:gz') as archive:
            archive.extractall(unarchive_dir)

        # It should contain `files_to_archive.json`
        fta_file = os.path.join(unarchive_dir, u'files_to_archive.json')
        assert os.path.exists(fta_file)

        # Which should properly contain { flattened_key -> original_filename }
        with open(fta_file) as fta:
            files_to_archive = json.loads(fta.read())

        assert files_to_archive == {
                u'model.text_field_embedder.token_embedders.elmo.options_file':
                        unicode(pathlib.Path(u'allennlp') / u'tests' / u'fixtures' / u'elmo' / u'options.json'),
                u'model.text_field_embedder.token_embedders.elmo.weight_file':
                        unicode(pathlib.Path(u'allennlp') / u'tests' / u'fixtures' / u'elmo' / u'lm_weights.hdf5'),
        }

        # Check that the unarchived contents of those files match the original contents.
        for key, original_filename in list(files_to_archive.items()):
            new_filename = os.path.join(unarchive_dir, u"fta", key)
            assert filecmp.cmp(original_filename, new_filename)

    def test_forward_works_with_projection_layer(self):
        params = Params({
                u'options_file': self.FIXTURES_ROOT / u'elmo' / u'options.json',
                u'weight_file': self.FIXTURES_ROOT / u'elmo' / u'lm_weights.hdf5',
                u'projection_dim': 20
                })
        word1 = [0] * 50
        word2 = [0] * 50
        word1[0] = 6
        word1[1] = 5
        word1[2] = 4
        word1[3] = 3
        word2[0] = 3
        word2[1] = 2
        word2[2] = 1
        word2[3] = 0
        embedding_layer = ElmoTokenEmbedder.from_params(vocab=None, params=params)
        input_tensor = torch.LongTensor([[word1, word2]])
        embedded = embedding_layer(input_tensor).data.numpy()
        assert embedded.shape == (1, 2, 20)

        input_tensor = torch.LongTensor([[[word1]]])
        embedded = embedding_layer(input_tensor).data.numpy()
        assert embedded.shape == (1, 1, 1, 20)




from __future__ import absolute_import
import re
from collections import defaultdict
#typing

#overrides
from unidecode import unidecode

from allennlp.data.tokenizers import Token
from allennlp.semparse.contexts.knowledge_graph import KnowledgeGraph
from io import open
try:
    from itertools import izip
except:
    izip = zip


DEFAULT_NUMBERS = [u'-1', u'0', u'1']
NUMBER_CHARACTERS = set([u'0', u'1', u'2', u'3', u'4', u'5', u'6', u'7', u'8', u'9', u'.', u'-'])
MONTH_NUMBERS = {
        u'january': 1,
        u'jan': 1,
        u'february': 2,
        u'feb': 2,
        u'march': 3,
        u'mar': 3,
        u'april': 4,
        u'apr': 4,
        u'may': 5,
        u'june': 6,
        u'jun': 6,
        u'july': 7,
        u'jul': 7,
        u'august': 8,
        u'aug': 8,
        u'september': 9,
        u'sep': 9,
        u'october': 10,
        u'oct': 10,
        u'november': 11,
        u'nov': 11,
        u'december': 12,
        u'dec': 12,
        }
ORDER_OF_MAGNITUDE_WORDS = {u'hundred': 100, u'thousand': 1000, u'million': 1000000}
NUMBER_WORDS = {
        u'zero': 0,
        u'one': 1,
        u'two': 2,
        u'three': 3,
        u'four': 4,
        u'five': 5,
        u'six': 6,
        u'seven': 7,
        u'eight': 8,
        u'nine': 9,
        u'ten': 10,
        u'first': 1,
        u'second': 2,
        u'third': 3,
        u'fourth': 4,
        u'fifth': 5,
        u'sixth': 6,
        u'seventh': 7,
        u'eighth': 8,
        u'ninth': 9,
        u'tenth': 10,
        }

class TableQuestionKnowledgeGraph(KnowledgeGraph):
    u"""
    A ``TableQuestionKnowledgeGraph`` represents the linkable entities in a table and a question
    about the table.  The linkable entities in a table are the cells and the columns of the table,
    and the linkable entities from the question are the numbers in the question.  We use the
    question to define our space of allowable numbers, because there are infinitely many numbers
    that we could include in our action space, and we really don't want to do that. Additionally, we
    have a method that returns the set of entities in the graph that are relevant to the question,
    and we keep the question for this method. See ``get_linked_agenda_items`` for more information.

    To represent the table as a graph, we make each cell and column a node in the graph, and
    consider a column's neighbors to be all cells in that column (and thus each cell has just one
    neighbor - the column it belongs to).  This is a rather simplistic view of the table. For
    example, we don't store the order of rows.

    We represent numbers as standalone nodes in the graph, without any neighbors.

    Additionally, when we encounter cells that can be split, we create ``fb:part.[something]``
    entities, also without any neighbors.
    """
    def __init__(self,
                 entities          ,
                 neighbors                      ,
                 entity_text                ,
                 question_tokens             )        :
        super(TableQuestionKnowledgeGraph, self).__init__(entities, neighbors, entity_text)
        self.question_tokens = question_tokens
        self._entity_prefixes = defaultdict(list)
        for entity, text in list(self.entity_text.items()):
            parts = text.split()
            if not parts:
                continue
            prefix = parts[0].lower()
            self._entity_prefixes[prefix].append(entity)

    @classmethod
    def read_from_file(cls, filename     , question             )                                 :
        u"""
        We read tables formatted as TSV files here. We assume the first line in the file is a tab
        separated list of column headers, and all subsequent lines are content rows. For example if
        the TSV file is:

        Nation      Olympics    Medals
        USA         1896        8
        China       1932        9

        we read "Nation", "Olympics" and "Medals" as column headers, "USA" and "China" as cells
        under the "Nation" column and so on.
        """
        return cls.read_from_lines(open(filename).readlines(), question)

    @classmethod
    def read_from_lines(cls, lines           , question             )                                 :
        cells = []
        # We assume the first row is column names.
        for row_index, line in enumerate(lines):
            line = line.rstrip(u'\n')
            if row_index == 0:
                columns = line.split(u'\t')
            else:
                cells.append(line.split(u'\t'))
        return cls.read_from_json({u"columns": columns, u"cells": cells, u"question": question})

    @classmethod
    def read_from_json(cls, json_object                )                                 :
        u"""
        We read tables formatted as JSON objects (dicts) here. This is useful when you are reading
        data from a demo. The expected format is::

            {"question": [token1, token2, ...],
             "columns": [column1, column2, ...],
             "cells": [[row1_cell1, row1_cell2, ...],
                       [row2_cell1, row2_cell2, ...],
                       ... ]}
        """
        entity_text                 = {}
        neighbors                              = defaultdict(list)

        # Getting number entities first.  Number entities don't have any neighbors, and their
        # "entity text" is the text from the question that evoked the number.
        question_tokens = json_object[u'question']
        for number, number_text in cls._get_numbers_from_tokens(question_tokens):
            entity_text[number] = number_text
            neighbors[number] = []
        for default_number in DEFAULT_NUMBERS:
            if default_number not in neighbors:
                neighbors[default_number] = []
                entity_text[default_number] = default_number

        # Following Sempre's convention for naming columns.  Sempre gives columns unique names when
        # columns normalize to a collision, so we keep track of these.  We do not give cell text
        # unique names, however, as `fb:cell.x` is actually a function that returns all cells that
        # have text that normalizes to "x".
        column_ids = []
        columns                 = {}
        for column_string in json_object[u'columns']:
            column_string = column_string.replace(u'\\n', u'\n')
            normalized_string = 'fb:row.row.{cls._normalize_string(column_string)}'
            if normalized_string in columns:
                columns[normalized_string] += 1
                normalized_string = '{normalized_string}_{columns[normalized_string]}'
            columns[normalized_string] = 1
            column_ids.append(normalized_string)
            entity_text[normalized_string] = column_string

        # Stores cell text to cell name, making sure that unique text maps to a unique name.
        cell_id_mapping                 = {}
        column_cells                  = [[] for _ in columns]
        for row_index, row_cells in enumerate(json_object[u'cells']):
            assert len(columns) == len(row_cells), (u"Invalid format. Row %d has %d cells, but header has %d"
                                                    u" columns" % (row_index, len(row_cells), len(columns)))
            # Following Sempre's convention for naming cells.
            row_cell_ids = []
            for column_index, cell_string in enumerate(row_cells):
                cell_string = cell_string.replace(u'\\n', u'\n')
                column_cells[column_index].append(cell_string)
                if cell_string in cell_id_mapping:
                    normalized_string = cell_id_mapping[cell_string]
                else:
                    base_normalized_string = 'fb:cell.{cls._normalize_string(cell_string)}'
                    normalized_string = base_normalized_string
                    attempt_number = 1
                    while normalized_string in list(cell_id_mapping.values()):
                        attempt_number += 1
                        normalized_string = "{base_normalized_string}_{attempt_number}"
                    cell_id_mapping[cell_string] = normalized_string
                row_cell_ids.append(normalized_string)
                entity_text[normalized_string] = cell_string
            for column_id, cell_id in izip(column_ids, row_cell_ids):
                neighbors[column_id].append(cell_id)
                neighbors[cell_id].append(column_id)

        for column in column_cells:
            if cls._should_split_column_cells(column):
                for cell_string in column:
                    for part_entity, part_string in cls._get_cell_parts(cell_string):
                        neighbors[part_entity] = []
                        entity_text[part_entity] = part_string
        return cls(set(neighbors.keys()), dict(neighbors), entity_text, question_tokens)

    @staticmethod
    def _normalize_string(string     )       :
        u"""
        These are the transformation rules used to normalize cell in column names in Sempre.  See
        ``edu.stanford.nlp.sempre.tables.StringNormalizationUtils.characterNormalize`` and
        ``edu.stanford.nlp.sempre.tables.TableTypeSystem.canonicalizeName``.  We reproduce those
        rules here to normalize and canonicalize cells and columns in the same way so that we can
        match them against constants in logical forms appropriately.
        """
        # Normalization rules from Sempre
        # \u201A -> ,

        return unidecode(string.lower())

    @staticmethod
    def _get_numbers_from_tokens(tokens             )                         :
        u"""
        Finds numbers in the input tokens and returns them as strings.  We do some simple heuristic
        number recognition, finding ordinals and cardinals expressed as text ("one", "first",
        etc.), as well as numerals ("7th", "3rd"), months (mapping "july" to 7), and units
        ("1ghz").

        We also handle year ranges expressed as decade or centuries ("1800s" or "1950s"), adding
        the endpoints of the range as possible numbers to generate.

        We return a list of tuples, where each tuple is the (number_string, token_text) for a
        number found in the input tokens.
        """
        numbers = []
        for i, token in enumerate(tokens):
            number                    = None
            token_text = token.text
            text = token.text.replace(u',', u'').lower()
            if text in NUMBER_WORDS:
                number = NUMBER_WORDS[text]

            magnitude = 1
            if i < len(tokens) - 1:
                next_token = tokens[i + 1].text.lower()
                if next_token in ORDER_OF_MAGNITUDE_WORDS:
                    magnitude = ORDER_OF_MAGNITUDE_WORDS[next_token]
                    token_text += u' ' + tokens[i + 1].text

            is_range = False
            if len(text) > 1 and text[-1] == u's' and text[-2] == u'0':
                is_range = True
                text = text[:-1]

            # We strip out any non-digit characters, to capture things like '7th', or '1ghz'.  The
            # way we're doing this could lead to false positives for something like '1e2', but
            # we'll take that risk.  It shouldn't be a big deal.
            text = u''.join(text[i] for i, char in enumerate(text) if char in NUMBER_CHARACTERS)

            try:
                # We'll use a check for float(text) to find numbers, because text.isdigit() doesn't
                # catch things like "-3" or "0.07".
                number = float(text)
            except ValueError:
                pass

            if number is not None:
                number = number * magnitude
                if u'.' in text:
                    number_string = u'%.3f' % number
                else:
                    number_string = u'%d' % number
                numbers.append((number_string, token_text))
                if is_range:
                    # TODO(mattg): both numbers in the range will have the same text, and so the
                    # linking score won't have any way to differentiate them...  We should figure
                    # out a better way to handle this.
                    num_zeros = 1
                    while text[-(num_zeros + 1)] == u'0':
                        num_zeros += 1
                    numbers.append((unicode(int(number + 10 ** num_zeros)), token_text))
        return numbers

    cell_part_regex = re.compile(r',\s|\n|/')
    @classmethod
    def _get_cell_parts(cls, cell_text     )                         :
        u"""
        Splits a cell into parts and returns the parts of the cell.  We return a list of
        ``(entity_name, entity_text)``, where ``entity_name`` is ``fb:part.[something]``, and
        ``entity_text`` is the text of the cell corresponding to that part.  For many cells, there
        is only one "part", and we return a list of length one.

        Note that you shouldn't call this on every cell in the table; SEMPRE decides to make these
        splits only when at least one of the cells in a column looks "splittable".  Only if you're
        splitting the cells in a column should you use this function.
        """
        parts = []
        for part_text in cls.cell_part_regex.split(cell_text):
            part_text = part_text.strip()
            part_entity = 'fb:part.{cls._normalize_string(part_text)}'
            parts.append((part_entity, part_text))
        return parts

    @classmethod
    def _should_split_column_cells(cls, column_cells           )        :
        u"""
        Returns true if there is any cell in this column that can be split.
        """
        return any(cls._should_split_cell(cell_text) for cell_text in column_cells)

    @classmethod
    def _should_split_cell(cls, cell_text     )        :
        u"""
        Checks whether the cell should be split.  We're just doing the same thing that SEMPRE did
        here.
        """
        if u', ' in cell_text or u'\n' in cell_text or u'/' in cell_text:
            return True
        return False

    def get_linked_agenda_items(self)             :
        u"""
        Returns entities that can be linked to spans in the question, that should be in the agenda,
        for training a coverage based semantic parser. This method essentially does a heuristic
        entity linking, to provide weak supervision for a learning to search parser.
        """
        agenda_items            = []
        for entity in self._get_longest_span_matching_entities():
            agenda_items.append(entity)
            # If the entity is a cell, we need to add the column to the agenda as well,
            # because the answer most likely involves getting the row with the cell.
            if u'fb:cell' in entity:
                agenda_items.append(self.neighbors[entity][0])
        return agenda_items

    def _get_longest_span_matching_entities(self):
        question = u" ".join([token.text for token in self.question_tokens])
        matches_starting_at                       = defaultdict(list)
        for index, token in enumerate(self.question_tokens):
            if token.text in self._entity_prefixes:
                for entity in self._entity_prefixes[token.text]:
                    if self.entity_text[entity].lower() in question:
                        matches_starting_at[index].append(entity)
        longest_matches            = []
        for index, matches in list(matches_starting_at.items()):
            longest_matches.append(sorted(matches, key=len)[-1])
        return longest_matches

    #overrides
    def __eq__(self, other):
        if isinstance(self, other.__class__):
            for key in self.__dict__:
                # We need to specially handle question tokens because they are Spacy's ``Token``
                # objects, and equality is not defined for them.
                if key == u"question_tokens":
                    self_tokens = self.__dict__[key]
                    other_tokens = other.__dict__[key]
                    if not all([token1.text == token2.text
                                for token1, token2 in izip(self_tokens, other_tokens)]):
                        return False
                else:
                    if not self.__dict__[key] == other.__dict__[key]:
                        return False
            return True
        return NotImplemented



# pylint: disable=no-self-use

from __future__ import absolute_import
from allennlp.common.testing import AllenNlpTestCase
from allennlp.semparse.type_declarations import type_declaration as base_types
from allennlp.semparse.type_declarations.type_declaration import (
        ComplexType,
        ANY_TYPE,
        )
from allennlp.semparse.type_declarations.wikitables_type_declaration import (
        ARG_EXTREME_TYPE,
        ArgExtremeType,
        CELL_TYPE,
        CONJUNCTION_TYPE,
        COUNT_TYPE,
        CountType,
        ROW_TYPE,
        REVERSE_TYPE,
        ReverseType,
        )


class WikiTablesTypeDeclarationTest(AllenNlpTestCase):
    def test_reverse_resolves_correctly(self):
        assert REVERSE_TYPE.resolve(CELL_TYPE) is None

        # Resolving against <?,<e,r>> should give <<r,e>,<e,r>>
        resolution = REVERSE_TYPE.resolve(ComplexType(ANY_TYPE, ComplexType(CELL_TYPE, ROW_TYPE)))
        assert resolution == ReverseType(ComplexType(ROW_TYPE, CELL_TYPE), ComplexType(CELL_TYPE, ROW_TYPE))

        # Resolving against <<r,?>,<e,?>> should give <<r,e>,<e,r>>
        resolution = REVERSE_TYPE.resolve(ComplexType(ComplexType(ROW_TYPE, ANY_TYPE),
                                                      ComplexType(CELL_TYPE, ANY_TYPE)))
        assert resolution == ReverseType(ComplexType(ROW_TYPE, CELL_TYPE), ComplexType(CELL_TYPE, ROW_TYPE))

        # Resolving against <<r,?>,?> should give <<r,?>,<?,r>>
        resolution = REVERSE_TYPE.resolve(ComplexType(ComplexType(ROW_TYPE, ANY_TYPE), ANY_TYPE))
        assert resolution == ReverseType(ComplexType(ROW_TYPE, ANY_TYPE), ComplexType(ANY_TYPE, ROW_TYPE))

        # Resolving against <<r,?>,<?,e>> should give None
        resolution = REVERSE_TYPE.resolve(ComplexType(ComplexType(ROW_TYPE, ANY_TYPE),
                                                      ComplexType(ANY_TYPE, CELL_TYPE)))
        assert resolution is None

    def test_conjunction_maps_to_correct_actions(self):
        valid_actions = base_types.get_valid_actions({u'and': u'O'},
                                                     {u'O': CONJUNCTION_TYPE},
                                                     set([CELL_TYPE]),
                                                     set([CELL_TYPE]))
        assert u'c -> [<#1,<#1,#1>>, c, c]' in valid_actions[u'c']

    def test_count_type_resolves_correctly(self):
        # Resolution should fail with basic type
        assert COUNT_TYPE.resolve(CELL_TYPE) is None

        # Resolution should fail when return type is not a number
        assert COUNT_TYPE.resolve(ComplexType(CELL_TYPE, ROW_TYPE)) is None

        # Resolution should resolve the return type to number
        assert COUNT_TYPE.resolve(ComplexType(CELL_TYPE, ANY_TYPE)) == CountType(CELL_TYPE)
        assert COUNT_TYPE.resolve(ComplexType(ANY_TYPE, ANY_TYPE)) == CountType(ANY_TYPE)

    def test_arg_extreme_type_resolves_correctly(self):
        # Resolution should fail on basic type
        assert ARG_EXTREME_TYPE.resolve(ROW_TYPE) is None

        assert ARG_EXTREME_TYPE.resolve(ComplexType(CELL_TYPE, ROW_TYPE)) is None

        other = ComplexType(ANY_TYPE,
                            ComplexType(ANY_TYPE,
                                        ComplexType(CELL_TYPE,
                                                    ComplexType(ComplexType(ANY_TYPE, CELL_TYPE),
                                                                CELL_TYPE))))
        resolution = ARG_EXTREME_TYPE.resolve(other)
        assert resolution == ArgExtremeType(CELL_TYPE)



# pylint: disable=no-self-use,invalid-name


from __future__ import division
from __future__ import absolute_import
from allennlp.common.testing import AllenNlpTestCase
from allennlp.models.archival import load_archive
from allennlp.predictors import Predictor


class TestSrlPredictor(AllenNlpTestCase):
    def test_uses_named_inputs(self):
        inputs = {
                u"sentence": u"The squirrel wrote a unit test to make sure its nuts worked as designed."
        }

        archive = load_archive(self.FIXTURES_ROOT / u'srl' / u'serialization' / u'model.tar.gz')
        predictor = Predictor.from_archive(archive, u'semantic-role-labeling')

        result = predictor.predict_json(inputs)

        words = result.get(u"words")
        assert words == [u"The", u"squirrel", u"wrote", u"a", u"unit", u"test",
                         u"to", u"make", u"sure", u"its", u"nuts", u"worked", u"as", u"designed", u"."]
        num_words = len(words)

        verbs = result.get(u"verbs")
        assert verbs is not None
        assert isinstance(verbs, list)

        assert any(v[u"verb"] == u"wrote" for v in verbs)
        assert any(v[u"verb"] == u"make" for v in verbs)
        assert any(v[u"verb"] == u"worked" for v in verbs)

        for verb in verbs:
            tags = verb.get(u"tags")
            assert tags is not None
            assert isinstance(tags, list)
            assert all(isinstance(tag, unicode) for tag in tags)
            assert len(tags) == num_words

    def test_batch_prediction(self):
        inputs = {
                u"sentence": u"The squirrel wrote a unit test to make sure its nuts worked as designed."
        }
        archive = load_archive(self.FIXTURES_ROOT / u'srl' / u'serialization' / u'model.tar.gz')
        predictor = Predictor.from_archive(archive, u'semantic-role-labeling')
        result = predictor.predict_batch_json([inputs, inputs])
        assert result[0] == result[1]

    def test_prediction_with_no_verbs(self):

        input1 = {u"sentence": u"Blah no verb sentence."}
        archive = load_archive(self.FIXTURES_ROOT / u'srl' / u'serialization' / u'model.tar.gz')
        predictor = Predictor.from_archive(archive, u'semantic-role-labeling')
        result = predictor.predict_json(input1)
        assert result == {u'words': [u'Blah', u'no', u'verb', u'sentence', u'.'], u'verbs': []}

        input2 = {u"sentence": u"This sentence has a verb."}
        results = predictor.predict_batch_json([input1, input2])
        assert results[0] == {u'words': [u'Blah', u'no', u'verb', u'sentence', u'.'], u'verbs': []}
        assert results[1] == {u'words': [u'This', u'sentence', u'has', u'a', u'verb', u'.'],
                              u'verbs': [{u'verb': u'has', u'description': u'This sentence has a verb .',
                                         u'tags': [u'O', u'O', u'O', u'O', u'O', u'O']}]}





from __future__ import division
from __future__ import absolute_import
#typing
#overrides

import torch

from allennlp.training.metrics.metric import Metric
try:
    from itertools import izip
except:
    izip = zip



class MentionRecall(Metric):
    def __init__(self)        :
        self._num_gold_mentions = 0
        self._num_recalled_mentions = 0

    #overrides
    def __call__(self,  # type: ignore
                 batched_top_spans              ,
                 batched_metadata                      ):
        for top_spans, metadata in izip(batched_top_spans.data.tolist(), batched_metadata):

            gold_mentions                       = set([mention for cluster in metadata[u"clusters"]
                                                   for mention in cluster])
            predicted_spans                       = set((span[0], span[1]) for span in top_spans)
            self._num_gold_mentions += len(gold_mentions)
            self._num_recalled_mentions += len(gold_mentions & predicted_spans)

    #overrides
    def get_metric(self, reset       = False)         :
        if self._num_gold_mentions == 0:
            recall = 0.0
        else:
            recall = self._num_recalled_mentions/float(self._num_gold_mentions)
        if reset:
            self.reset()
        return recall

    #overrides
    def reset(self):
        self._num_gold_mentions = 0
        self._num_recalled_mentions = 0

MentionRecall = Metric.register(u"mention_recall")(MentionRecall)



u"""
This module contains code relating to processing logical forms for semantic parsing.  The main
functionality provided by this code is:

    1. A way of declaring a type system for use in processing logical forms.  This determines which
    (lisp-like) logical forms are valid in your logical form language, and it enables us to parse
    statements in that language into logical expressions.  We rely heavily on NLTK for the logic
    processing here.  This is the ``type_declarations`` module.

    2. A way of specifying extra "context" that's available to the parser / type system.
    Currently, this just consists of a ``KnowledgeGraph`` that can represent table entities, or
    question-specific entities.  These entities can be made available to the type system to allow
    context-specific production rules.  This is the ``contexts`` module.

    3. A way to bundle the two things above together in the context of a particular training /
    testing instance.  This bundling is called a ``World``, and this is the primary means that a
    model should use to interact with the type system.  The ``World`` has methods to get all
    possible actions in any state, all context entities that need to be linked, and so on.  This is
    the ``worlds`` module.

Note that the main reason you would use this code is to get the set of actions that are available
at any point during constrained decoding.  If you have some other means of doing that, you might
not need this code at all.
"""

# We have some circular dependencies with the data code.  Importing this loads all of the data code
# before we try to import stuff under semparse.  This means that we can't do
# `from allennlp.semparse import [whatever]` in the `data` module, but it resolves the other
# dependency issues.  If you want to import semparse stuff from the data code, just use a more
# complete path, like `from allennlp.semparse.worlds import WikiTablesWorld`.
# from allennlp.data.tokenizers import Token as _
# from allennlp.semparse.worlds.world import ParsingError, World
# from allennlp.semparse.action_space_walker import ActionSpaceWalker



# pylint: disable=no-self-use,invalid-name


from __future__ import division
from __future__ import absolute_import
import pytest

from allennlp.data.dataset_readers import StanfordSentimentTreeBankDatasetReader
from allennlp.common import Params
from allennlp.common.util import ensure_list
from allennlp.common.testing import AllenNlpTestCase

class TestStanfordSentimentTreebankReader(object):
    sst_path = AllenNlpTestCase.FIXTURES_ROOT / u"data" / u"sst.txt"

    @pytest.mark.parametrize(u"lazy", (True, False))
    def test_read_from_file(self, lazy):
        reader = StanfordSentimentTreeBankDatasetReader(lazy=lazy)
        instances = reader.read(self.sst_path)
        instances = ensure_list(instances)

        instance1 = {u"tokens": [u"The", u"actors", u"are", u"fantastic", u"."],
                     u"label": u"4"}
        instance2 = {u"tokens": [u"It", u"was", u"terrible", u"."],
                     u"label": u"0"}
        instance3 = {u"tokens": [u"Chomp", u"chomp", u"!"],
                     u"label": u"2"}

        assert len(instances) == 3
        fields = instances[0].fields
        assert [t.text for t in fields[u"tokens"].tokens] == instance1[u"tokens"]
        assert fields[u"label"].label == instance1[u"label"]
        fields = instances[1].fields
        assert [t.text for t in fields[u"tokens"].tokens] == instance2[u"tokens"]
        assert fields[u"label"].label == instance2[u"label"]
        fields = instances[2].fields
        assert [t.text for t in fields[u"tokens"].tokens] == instance3[u"tokens"]
        assert fields[u"label"].label == instance3[u"label"]

    def test_use_subtrees(self):
        reader = StanfordSentimentTreeBankDatasetReader(use_subtrees=True)
        instances = reader.read(self.sst_path)
        instances = ensure_list(instances)

        instance1 = {u"tokens": [u"The", u"actors", u"are", u"fantastic", u"."],
                     u"label": u"4"}
        instance2 = {u"tokens": [u"The", u"actors"],
                     u"label": u"2"}
        instance3 = {u"tokens": [u"The"],
                     u"label": u"2"}

        assert len(instances) == 21
        fields = instances[0].fields
        assert [t.text for t in fields[u"tokens"].tokens] == instance1[u"tokens"]
        assert fields[u"label"].label == instance1[u"label"]
        fields = instances[1].fields
        assert [t.text for t in fields[u"tokens"].tokens] == instance2[u"tokens"]
        assert fields[u"label"].label == instance2[u"label"]
        fields = instances[2].fields
        assert [t.text for t in fields[u"tokens"].tokens] == instance3[u"tokens"]
        assert fields[u"label"].label == instance3[u"label"]

    def test_3_class(self):
        reader = StanfordSentimentTreeBankDatasetReader(granularity=u"3-class")
        instances = reader.read(self.sst_path)
        instances = ensure_list(instances)

        instance1 = {u"tokens": [u"The", u"actors", u"are", u"fantastic", u"."],
                     u"label": u"2"}
        instance2 = {u"tokens": [u"It", u"was", u"terrible", u"."],
                     u"label": u"0"}
        instance3 = {u"tokens": [u"Chomp", u"chomp", u"!"],
                     u"label": u"1"}

        assert len(instances) == 3
        fields = instances[0].fields
        assert [t.text for t in fields[u"tokens"].tokens] == instance1[u"tokens"]
        assert fields[u"label"].label == instance1[u"label"]
        fields = instances[1].fields
        assert [t.text for t in fields[u"tokens"].tokens] == instance2[u"tokens"]
        assert fields[u"label"].label == instance2[u"label"]
        fields = instances[2].fields
        assert [t.text for t in fields[u"tokens"].tokens] == instance3[u"tokens"]
        assert fields[u"label"].label == instance3[u"label"]

    def test_2_class(self):
        reader = StanfordSentimentTreeBankDatasetReader(granularity=u"2-class")
        instances = reader.read(self.sst_path)
        instances = ensure_list(instances)

        instance1 = {u"tokens": [u"The", u"actors", u"are", u"fantastic", u"."],
                     u"label": u"1"}
        instance2 = {u"tokens": [u"It", u"was", u"terrible", u"."],
                     u"label": u"0"}

        assert len(instances) == 2
        fields = instances[0].fields
        assert [t.text for t in fields[u"tokens"].tokens] == instance1[u"tokens"]
        assert fields[u"label"].label == instance1[u"label"]
        fields = instances[1].fields
        assert [t.text for t in fields[u"tokens"].tokens] == instance2[u"tokens"]
        assert fields[u"label"].label == instance2[u"label"]

    def test_from_params(self):
        # pylint: disable=protected-access
        params = Params({u"use_subtrees": True, u"granularity": u"5-class"})
        reader = StanfordSentimentTreeBankDatasetReader.from_params(params)
        assert reader._use_subtrees is True
        assert reader._granularity == u"5-class"



# pylint: disable=no-self-use,invalid-name

from __future__ import absolute_import
from numpy.testing import assert_almost_equal
import torch
from torch.nn import Embedding, Parameter

from allennlp.common.testing import AllenNlpTestCase
from allennlp.modules import TimeDistributed


class TestTimeDistributed(AllenNlpTestCase):
    def test_time_distributed_reshapes_correctly(self):
        char_embedding = Embedding(2, 2)
        char_embedding.weight = Parameter(torch.FloatTensor([[.4, .4], [.5, .5]]))
        distributed_embedding = TimeDistributed(char_embedding)
        char_input = torch.LongTensor([[[1, 0], [1, 1]]])
        output = distributed_embedding(char_input)
        assert_almost_equal(output.data.numpy(),
                            [[[[.5, .5], [.4, .4]], [[.5, .5,], [.5, .5]]]])

    def test_time_distributed_works_with_multiple_inputs(self):
        module = lambda x, y: x + y
        distributed = TimeDistributed(module)
        x_input = torch.LongTensor([[[1, 2], [3, 4]]])
        y_input = torch.LongTensor([[[4, 2], [9, 1]]])
        output = distributed(x_input, y_input)
        assert_almost_equal(output.data.numpy(), [[[5, 4], [12, 5]]])
